@@ -17,7 +17,6 @@ This module contains builtin handlers for events emitted by botocore.
 """
 
 import base64
-import hashlib
 import logging
 import xml.etree.cElementTree
 import copy
@@ -25,7 +24,7 @@ import re
 import warnings
 
 from botocore.compat import urlsplit, urlunsplit, unquote, \
-    json, quote, six, unquote_str
+    json, quote, six, unquote_str, ensure_bytes, get_md5, MD5_AVAILABLE
 from botocore.docs.utils import AutoPopulatedParam
 from botocore.docs.utils import HideParamFromOperations
 from botocore.docs.utils import AppendParamDocumentation
@@ -134,13 +133,13 @@ def calculate_md5(params, **kwargs):
 
 
 def _calculate_md5_from_bytes(body_bytes):
-    md5 = hashlib.md5(body_bytes)
+    md5 = get_md5(body_bytes)
     return md5.digest()
 
 
 def _calculate_md5_from_file(fileobj):
     start_position = fileobj.tell()
-    md5 = hashlib.md5()
+    md5 = get_md5()
     for chunk in iter(lambda: fileobj.read(1024 * 1024), b''):
         md5.update(chunk)
     fileobj.seek(start_position)
@@ -150,7 +149,7 @@ def _calculate_md5_from_file(fileobj):
 def conditionally_calculate_md5(params, **kwargs):
     """Only add a Content-MD5 when not using sigv4"""
     signer = kwargs['request_signer']
-    if signer.signature_version != 'v4':
+    if signer.signature_version not in ['v4', 's3v4'] and MD5_AVAILABLE:
         calculate_md5(params, **kwargs)
 
 
@@ -188,13 +187,14 @@ def copy_source_sse_md5(params, **kwargs):
 def _sse_md5(params, sse_member_prefix='SSECustomer'):
     if not _needs_s3_sse_customization(params, sse_member_prefix):
         return
+
     sse_key_member = sse_member_prefix + 'Key'
     sse_md5_member = sse_member_prefix + 'KeyMD5'
     key_as_bytes = params[sse_key_member]
     if isinstance(key_as_bytes, six.text_type):
         key_as_bytes = key_as_bytes.encode('utf-8')
     key_md5_str = base64.b64encode(
-        hashlib.md5(key_as_bytes).digest()).decode('utf-8')
+        get_md5(key_as_bytes).digest()).decode('utf-8')
     key_b64_encoded = base64.b64encode(key_as_bytes).decode('utf-8')
     params[sse_key_member] = key_b64_encoded
     params[sse_md5_member] = key_md5_str
@@ -600,10 +600,32 @@ def decode_list_object(parsed, context, **kwargs):
     # This is needed because we are passing url as the encoding type. Since the
     # paginator is based on the key, we need to handle it before it can be
     # round tripped.
-    if 'Contents' in parsed and parsed.get('EncodingType') == 'url' and \
+    #
+    # From the documentation: If you specify encoding-type request parameter,
+    # Amazon S3 includes this element in the response, and returns encoded key
+    # name values in the following response elements:
+    # Delimiter, Marker, Prefix, NextMarker, Key.
+    if parsed.get('EncodingType') == 'url' and \
                     context.get('EncodingTypeAutoSet'):
-        for content in parsed['Contents']:
-            content['Key'] = unquote_str(content['Key'])
+        # URL decode top-level keys in the response if present.
+        top_level_keys = ['Delimiter', 'Marker', 'NextMarker']
+        for key in top_level_keys:
+            if key in parsed:
+                parsed[key] = unquote_str(parsed[key])
+        # URL decode nested keys from the response if present.
+        nested_keys = [('Contents', 'Key'), ('CommonPrefixes', 'Prefix')]
+        for (top_key, child_key) in nested_keys:
+            if top_key in parsed:
+                for member in parsed[top_key]:
+                    member[child_key] = unquote_str(member[child_key])
+
+
+def convert_body_to_file_like_object(params, **kwargs):
+    if 'Body' in params:
+        if isinstance(params['Body'], six.string_types):
+            params['Body'] = six.BytesIO(ensure_bytes(params['Body']))
+        elif isinstance(params['Body'], six.binary_type):
+            params['Body'] = six.BytesIO(params['Body'])
 
 
 # This is a list of (event_name, handler).
@@ -611,6 +633,10 @@ def decode_list_object(parsed, context, **kwargs):
 # automatically registered with that Session.
 
 BUILTIN_HANDLERS = [
+    ('before-parameter-build.s3.UploadPart',
+     convert_body_to_file_like_object, REGISTER_LAST),
+    ('before-parameter-build.s3.PutObject',
+     convert_body_to_file_like_object, REGISTER_LAST),
     ('creating-client-class', add_generate_presigned_url),
     ('creating-client-class.s3', add_generate_presigned_post),
     ('creating-client-class.iot-data', check_openssl_supports_tls_version_1_2),
